/*
 * HTTPARSER.H
 * Contains templates and parser fuctions.
 *
 * Copyright (C) 2008  Acri Emanuele <crossbower@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __HTTPARSER_H__
#define __HTTPARSER_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <regex.h>
#include <getopt.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define VERSION "v0.6"
#define AUTHOR "Acri Emanuele (crossbower@gmail.com)"

#define MIN_ARG 256
#define MAX_ARG 1024

#define MAX_DATA 2048


/*
 * IP range structure
 */
struct iprange {
    char range[MIN_ARG];    // ip range
    char block[MIN_ARG];    // block
    char addr[MIN_ARG];     // address
    
    char http_request[MAX_DATA]; // http request
    int reqtype;   // request type
    
    int start;     // fist host
    int end;       // last host
    int curr;      // current host
    
    int port;      // port
    int time;      // child lifetime
    
    int ipver;     // ip version
    
    int verbose:1;     // verbose
    int cookiejar:1;   // cookie-jar output
    int fingerprint:1; // fingerprint
};
#define VER_IPV4 4
#define VER_IPV6 6

#define REQ_GET     1
#define REQ_POST    2
#define REQ_HEAD    3
#define REQ_OPTIONS 4
#define REQ_PUT     5
#define REQ_DELETE  6
#define REQ_TRACE   7
#define REQ_CONNECT 8

/*
 * Regular expression used by
 * parseopt() and parseipr() functions
 */
char *reg_ipv4_range =
    "^[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}[-][0-9]{1,3}$";

char *reg_ipv4_addr =
    "^[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}$";

char *reg_ipv6_range =
    "^[0-9a-fA-F:]{1,42}[-][0-9a-fA-F]{1,4}$";

char *reg_ipv6_addr =
    "^[0-9a-fA-F:]{1,42}$";

char *reg_number =
    "\\d+?";

/*
 * Http 1.1 requests
 */
char *get_request =
"GET / HTTP/1.1\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define GET_REQ 1

char *head_request =
"HEAD / HTTP/1.1\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define HEAD_REQ 2

char *get_not_exist_request =
"GET /foo.foo HTTP/1.1\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define GET_NOT_REQ 3

char *get_wrong_version_request =
"GET / HTTP/9.9\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define GET_WRONG_REQ 4

char *options_request =
"OPTIONS * HTTP/1.1\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define OPTIONS_REQ 5

/*
char *put_request =
"PUT / HTTP/1.1\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define PUT_REQ 6
*/

char *delete_request =
"DELETE / HTTP/1.1\r\n" \
"Host: localhost\r\n" \
"\r\n";
#define DELETE_REQ 7

/*
 * Program usage
 */
void usage(FILE *out, char *error)
{
    char usage[] =
        "\nHTTSquash scanner " VERSION " - " AUTHOR "\n" \
        "Usage:\n" \
        "  httsquash -r range [options]\n" \
        "Options:\n" \
        "  -r\trange of ip addresses or target dns name\n" \
        "    \texamples: 208.67.1.1-254, 2001::1428:57ab-6344, google.com\n" \
        "  -p\tport (default 80)\n" \
        "Extra options:\n" \
        "  -t\ttime in seconds (default 3)\n" \
        "  -v\tfull answer (include html data)\n" \
        "  -j\tcookie jar separator (\"%%\")\n" \
        "  -P\tpersonalized http request, payload file\n" \
        "  -T\trequest type (default get)\n" \
        "    \tsupported types: get, head, get_not (not exist),\n" \
        "    \t                 get_wrong (wrong version), options,\n" \
        "    \t                 delete\n" \
        "  -F\tenable fingerprinting (it uses request type...)\n";

    fputs(usage, out);

    if(error)
        fputs(error, out);

    exit(1);
}

/*
 * A personalized strncat
 */
char * myncat(char *dest, const char *src, size_t n)
{
    size_t dest_len = strlen(dest);
    size_t i;

    for (i = 0 ; i < (n-dest_len-1) && src[i] != '\0' ; i++) {
        dest[dest_len + i] = src[i];
    }
    
    dest[dest_len + i] = '\0';

    return dest;
}


/*
 * Regular expression matching function
 * Returns:
 *    -1  error on pattern compilation
 *     0  don't match
 *     1  match
 */
int regmatch(const char* str, char* regex) {
    regex_t re;

    // Compile the regex pattern.
    if (regcomp(&re, regex, REG_EXTENDED | REG_NOSUB) != 0) {
        fprintf(stderr, "Error on regex pattern compilation.\n");
        return -1;
    }

    // Use the pattern on the passed string
    if (regexec(&re, str, 0, NULL, 0))  {
        return 0;
    }

    regfree(&re);
    return 1;
}

/*
 * Read http request from a file
 */
int readhttprequest(struct iprange *ipr, char *path) {
    FILE *file;

    if((file=fopen(path, "r"))==NULL)
        usage(stderr, "\nError opening file for read the http request.\n");

    if(!fread(ipr->http_request, 1, MAX_DATA-1, file))
        usage(stderr, "\nError reading http request from file.\n");

    fclose(file);
    
    return 1;
}

/*
 * Parser for command line options
 * See getopt(3)...
 */
int parseopt(struct iprange *ipr, int argc, char **argv)
{
    char *c, *x;
    int i=0;
    char ch;
    
    char range[MIN_ARG],  // ip range
         *request=NULL;   // http request
    
    int start = 0,       // fist host
        end  = 0,        // last host
        curr = 0,        // current host
        port = 80,       // port
        time = 3,        // child lifetime
        verbose = 0,     // verbose
        cookiejar = 0,   // cookie-jar output
        reqtype = 1,     // request type
        fingerprint = 0; // fingerprint
    
    int ipver=0;   // ip version
    int single=0;  // single ip?
    
    struct addrinfo hints, *res=NULL;
    void *addr=NULL;
    
    // cleaning
    range[0]='\0';
    memset(ipr, 0, sizeof(struct iprange));
    memset(&hints, 0, sizeof(hints));
    
    char *shortopt = "r:p:t:vjP:T:F"; // short options
    
    while ((ch = getopt (argc, argv, shortopt)) != -1)
        switch (ch) {
        
            case 'r': // IP range or dns name
                
                // ipv4 range
                if(regmatch(optarg, reg_ipv4_range)) {
                    ipver=VER_IPV4;
                    strncpy(range, optarg, MIN_ARG-1);
                }
                
                // ipv6 range
                else if(regmatch(optarg, reg_ipv6_range)) {
                    ipver=VER_IPV6;
                    strncpy(range, optarg, MIN_ARG-1);
                }
                
                // dns name
                else if(getaddrinfo(optarg, NULL, &hints, &res) == 0) {
                
                    if (res->ai_family == AF_INET) { // IPv4
                        struct sockaddr_in *ipv4 = (struct sockaddr_in *) res->ai_addr;
                        addr = &(ipv4->sin_addr);
                        ipver=VER_IPV4;
                    }
                    
                    else if (res->ai_family == AF_INET6) { // IPv6
                        struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *)res->ai_addr;
                        addr = &(ipv6->sin6_addr);
                        ipver=VER_IPV6;
                    }

                    // convert the IP to a string and print it:
                    inet_ntop(res->ai_family, addr, range, MIN_ARG-1);
                    
                    single=1;
                    freeaddrinfo(res);
                }
                
                else {
                    usage(stderr, "\nArgument range is invalid.\n");
                }

                break;
            
            case 'p': // Port
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument port is invalid.\n");
                port = atoi(optarg);
                break;
            
            case 't': // Time
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument time is invalid.\n");
                time = atoi(optarg);
                break;
                
            case 'v': // Full answer
                verbose = 1;
                break;
            
            case 'j': // Cookie jar
                cookiejar = 1;
                break;
            
            case 'P': // Payload file
                request = optarg;
                break;
            
            case 'T': // Request type
                if(!strcasecmp(optarg, "get")) {
                    reqtype = GET_REQ;
                }
                else if(!strcasecmp(optarg, "head")) {
                    reqtype = HEAD_REQ;
                }
                else if(!strcasecmp(optarg, "get_not")) {
                    reqtype = GET_NOT_REQ;
                }
                else if(!strcasecmp(optarg, "get_wrong")) {
                    reqtype = GET_WRONG_REQ;
                }
                else if(!strcasecmp(optarg, "options")) {
                    reqtype = OPTIONS_REQ;
                }
                /*else if(!strcasecmp(optarg, "put")) {
                    reqtype = PUT_REQ;
                }*/
                else if(!strcasecmp(optarg, "delete")) {
                    reqtype = DELETE_REQ;
                }
                else {
                    usage(stderr, "\nArgument request type is invalid.\n");
                }
                break;
            
            case 'F': // Enable fingerprinting
                fingerprint = 1;
                break;
                            
            case '?':
            default:
                usage(stderr, NULL);
        }


    /* Check collected args */
    if(range[0]=='\0') {
        usage(stderr, "\nWhere is the range? Uff...\n");
    }
    
    if(time<1) time = 3;
    
    if(port<1) port = 80;
    
    
    /* Fill iprange structure */
    
    // port
    ipr->port=port;
    
    // time
    ipr->time=time;    
    
    // verbose
    ipr->verbose=verbose;
    
    // cookiejar
    ipr->cookiejar=cookiejar;
    
    // request type
    ipr->reqtype=reqtype;
    
    // fingerprint
    ipr->fingerprint=fingerprint;
    
    // ipversion
    ipr->ipver=ipver;

    // range
	if(range[0]!='\0') {
	
        strncpy(ipr->range, range, MIN_ARG-1);
        
        // ipv4
        if(ipver==VER_IPV4) {
            c = range;
            x = range;

            while(*c!='\0') c++;
            while(*c!='.') c--;

            *c = '\0';
            strncpy(ipr->block, range, MIN_ARG-1);
            *c = '.';

            // start
            if(!single) {
                while(*x!='-') x++;
                *x='\0';
            }
            
            c++;
            
            ipr->start=atoi(c);
    
            // single ip
            if(single) {
                strcat(ipr->range,"-");
                strcat(ipr->range,c);
                ipr->end=ipr->start;
            }
            
            // range
            else {
                x++;
                ipr->end=atoi(x);
            }
        }
        
        // ipv6
        else if(ipver==VER_IPV6) {
            c = range;
            x = range;

            while(*c!='\0') c++;
            while(*c!=':') c--;

            *c = '\0';
            strncpy(ipr->block, range, MIN_ARG-1);
            *c = ':';

            // start
            if(!single) {
                while(*x!='-') x++;
                *x='\0';
            }
            
            c++;
            
            sscanf(c, "%x", &ipr->start);

            // single ip
            if(single) {
                strcat(ipr->range,"-");
                strcat(ipr->range,c);
                ipr->end=ipr->start;
            }
            
            // range
            else {
                x++;
                sscanf(x, "%x", &ipr->end);
            }
        }
        
        // current
        ipr->curr = ipr->start;
	}
   
    // check range
    if(ipr->end < ipr->start)
        usage(stderr, "\nArgument ip-range is impossible ;)\n");
    
    //http_request
    if(request) {
        readhttprequest(ipr, request);
    }
    else {
        if(reqtype == GET_REQ) {
            strncpy(ipr->http_request, get_request, MAX_DATA-1);
        }
        else if(reqtype == HEAD_REQ) {
            strncpy(ipr->http_request, head_request, MAX_DATA-1);
        }
        else if(reqtype == GET_NOT_REQ) {
            strncpy(ipr->http_request, get_not_exist_request, MAX_DATA-1);
        }
        else if(reqtype == GET_WRONG_REQ) {
            strncpy(ipr->http_request, get_wrong_version_request, MAX_DATA-1);
        }
        else if(reqtype == OPTIONS_REQ) {
            strncpy(ipr->http_request, options_request, MAX_DATA-1);
        }
        /*else if(reqtype == PUT_REQ) {
            strncpy(ipr->http_request, put_request, MAX_DATA-1);
        }*/
        else if(reqtype == DELETE_REQ) {
            strncpy(ipr->http_request, delete_request, MAX_DATA-1);
        }
    }
    
    return 1;
}

#endif /* __HTTPARSER_H__ */

