/*
 * ARGPARSER.H
 * Usage and parser...
 *
 * Copyright (C) 2008  Acri Emanuele <crossbower@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __ARGPARSER_H__
#define __ARGPARSER_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <getopt.h>
#include <regex.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <time.h>

#define VERSION "v0.6"
#define AUTHOR "Acri Emanuele (crossbower@gmail.com)"

#define MIN_ARG 128
#define MAX_ARG 2048
#define MAX_PAY 15

#define IPTABLES_ENABLE  "iptables -A OUTPUT -p tcp --tcp-flags ALL RST -j DROP"
#define IPTABLES_DISABLE "iptables -D OUTPUT -p tcp --tcp-flags ALL RST -j DROP"

#define BLACKHOLE_ENABLE  "sysctl net.inet.tcp.blackhole=2"
#define BLACKHOLE_DISABLE "sysctl net.inet.tcp.blackhole=0"

/*
 * Stage struct (payload)
 */
struct stage {
    int active;
    int type;
    int action;
    char payload[MAX_ARG*4];
};

#define TYPE_ACK 1
#define TYPE_ACK_PUSH 2
#define TYPE_ACTION 3

#define ACT_WAIT_REPLY 1
#define ACT_EXIT 2
#define ACT_RESET 3

/*
 * Options structure
 */
struct options {
    int dport;  // destination port
    int xport;  // first source port
    int yport;  // last source port
    int time;   // sleep time
    int alarm;  // alarm time
    int window; // tcp window size
    int frag;   // fragment offset
    int fragc;  // fragment counter
    
    int ipver;  // ip version
    
    int syn:1;  // only syn flood
    int ack:1;  // send acknowledgment packets
    int fin:1;  // send finalize packets
    int rst:1;  // send reset packets
    int mlt:1;  // multistage payload
    int loop:1; // infinite loop
    
    int children; // max processes to spawn
    int verbose;  // verbosity
    
    char dstaddr[MIN_ARG]; // destination address
    char srcaddr[MIN_ARG]; // source address
    char device[MIN_ARG];  // network interface
    
    char firewall[2][MAX_ARG]; // firewall rules
    
    char payload[MAX_ARG*4];
    
    struct stage *currstg;
    struct stage stg[MAX_PAY]; // array of stage structs
};
#define VER_IPV4 4
#define VER_IPV6 6


/*
 * Regular expression used by
 * parseopt() function
 */
 char *reg_ipv4_range =
    "^[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}[-][0-9]{1,3}$";

char *reg_ipv4_addr =
    "^[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}$";

char *reg_ipv6_range =
    "^[0-9a-fA-F:]{1,42}[-][0-9a-fA-F]{1,4}$";

char *reg_ipv6_addr =
    "^[0-9a-fA-F:]{1,42}$";

char *reg_payload_head =
    "^\\.PAYHEAD (ACK|PUSHACK)\\s{0,}$";

char *reg_payload_tail =
    "^\\.PAYTAIL (REPL|EXIT|RESET)\\s{0,}$";

char *reg_action =
    "^\\.ACTION (REPL|EXIT|RESET)\\s{0,}$";

char *reg_number =
    "\\d+?";
    
/*
 * Program usage
 */
void usage(FILE *out, char *error)
{
    char usage[] =
        "\nLetDown 3wh+payload flooder " VERSION " - " AUTHOR "\n" \
        "Usage:\n" \
        "  letdown -d destination ip -p port [options]\n" \
        "Options:\n" \
        "  -d\tdestination ip address or dns name, target\n" \
        "  -p\tdestination port\n" \
        "  -s\tsource ip address\n" \
        "  -x\tfirst source port (default 1025)\n" \
        "  -y\tlast source port (default 65534)\n" \
        "  -l\tenables infinite loop mode\n"
        "  -i\tnetwork interface\n" \
        "  -t\tsleep time in microseconds (default 10000)\n" \
        "  -a\tmax time in second for waiting responses (default 40)\n" \
        "  -m\tmax processes to spawn (default 8)\n"
        "Extra options:\n" \
        "  -v\tverbosity level (0=quiet, 1=normal, 2=verbose)\n"
        "  -f\tautomagically set firewall rules for blocking\n" \
        "    \trst packet generated by the kernel\n"
        "    \texamples: -f iptables, -f blackhole (for freebsd)\n" \
        "  -L\tinteraction level with the target (packets to send)\n" \
        "    \ts  syn flooding, no 3-way-handshake\n" \
        "    \ta  send acknowledgment packets (polite mode)\n" \
        "    \tf  send finalize packets\n" \
        "    \tr  send reset packets (check firewall rules...)\n" \
        "  -W\twindow size for ack packets (ex: 0-window attack)\n" \
        "  -O\tenable ack fragmentation and set fragment offset delta\n" \
        "  -C\tfragment counter if fragmentation is enabled (default 1)\n" \
        "  -P\tpayload file (see payloads directory...)\n" \
        "  -M\tmultistage payload file (see payloads directory...)\n";

    fputs(usage, out);

    if(error)
        fputs(error, out);

    exit(1);
}

/*
 * Regular expression matching function
 * Returns:
 *    -1  error on pattern compilation
 *     0  don't match
 *     1  match
 */
int regmatch(const char* str, char* regex)
{
    regex_t re;

    // Compile the regex pattern.
    if (regcomp(&re, regex, REG_EXTENDED | REG_NOSUB) != 0) {
        fprintf(stderr, "Error on regex pattern compilation.\n");
        return -1;
    }

    // Use the pattern on the passed string
    if (regexec(&re, str, 0, NULL, 0))  {
        return 0;
    }

    regfree(&re);
    return 1;
}

/*
 * Check verbosity
 */
int verbosity(int verbose, struct options *opt)
{
    if(verbose > opt->verbose)
        return 0;
        
    return 1;
}

/*
 * Print time
 */
void printtime()
{
  struct tm *local;
  time_t t;

  t = time(NULL);
  local = localtime(&t);
  printf("%s", asctime(local));
}

/*
 * Get payload from a file
 */
int getpayload(struct options *opt, char *path) {
    FILE *file;

    // open file
    if((file=fopen(path, "r"))==NULL)
        usage(stderr, "\nError opening payload file.\n");

    if(!fread(opt->payload, 1, MAX_ARG*4, file))
        usage(stderr, "\nError reading  payload from file.\n");

    fclose(file);
    
    return 1;
}

/*
 * Get multistage payload from a file
 */
int getmultipayload(struct options *opt, char *path) {
    FILE *file;
    char buffer[MAX_ARG*4];
    int i=0, size=0, tot=0;
    
    /*
     * Multistage payload fields:
     * 
     * ##ACTION type
     * ##PAYHEAD type
     * ##PAYTAIL action
     */
    char headack[] = "^\\.PAYHEAD ACK\\s{0,}$";
    char headackpush[] = "^\\.PAYHEAD PUSHACK\\s{0,}$";
    
    char tailrepl[]  = "^\\.PAYTAIL REPL\\s{0,}$";
    char tailexit[]  = "^\\.PAYTAIL EXIT\\s{0,}$";
    char tailreset[] = "^\\.PAYTAIL RESET\\s{0,}$";
    
    char actrepl[]  = "^\\.ACTION REPL\\s{0,}$";
    char actexit[]  = "^\\.ACTION EXIT\\s{0,}$";
    char actreset[] = "^\\.ACTION RESET\\s{0,}$";

    // open file
    if((file=fopen(path, "r"))==NULL)
        usage(stderr, "\nError opening payload file.\n");
    
    buffer[0]='\0';
    
    while(i<(MAX_PAY-1) && fgets(buffer, sizeof(buffer), file)!=NULL) {
    
        // ignore comments
        if(buffer[0]=='#') {
            continue;
        }
        
        // search action
        else if(regmatch(buffer, reg_action)) {
            
            opt->stg[i].type = TYPE_ACTION;
            
            // parse action
            if (regmatch(buffer, actrepl)) {
                opt->stg[i].action = ACT_WAIT_REPLY;
            }
            else if (regmatch(buffer, actexit)) {
                opt->stg[i].action = ACT_EXIT;
            }
            else {
                opt->stg[i].action = ACT_RESET;
            }
            
            // struct stg[i] is ready...
            opt->stg[i].active = 1;
            i++;
        }
        
        // search payload header
        else if(regmatch(buffer, reg_payload_head)) {
            
            // parse head
            if(regmatch(buffer, headack)) {
                opt->stg[i].type = TYPE_ACK;
            }
            else {
                opt->stg[i].type = TYPE_ACK_PUSH;
            }
            
            // read payload
            while((size = (int) fgets(buffer, sizeof(buffer), file))!=0) {

                // search payload tail
                if(regmatch(buffer, reg_payload_tail)) {
                
                    // parse tail
                    if(regmatch(buffer, tailrepl)) {
                        opt->stg[i].action = ACT_WAIT_REPLY;
                    }
                    else if(regmatch(buffer, tailexit)) {
                        opt->stg[i].action = ACT_EXIT;
                    }
                    else {
                        opt->stg[i].action = ACT_RESET;
                    }
                  
                    break;
                }
                
                // fill payload
                tot+=size;
                if(tot < MAX_ARG*4) {
                    strcat(opt->stg[i].payload, buffer);
                }
            }
            
            // struct stg[i] is ready...
            opt->stg[i].active = 1;
            i++;
            tot=0;
        }
        
        buffer[0]='\0';
    }
    
    // current stage point at first stage
    opt->currstg = &opt->stg[0];

    fclose(file);
    
    return 1;
}


/*
 * Parser for command line options
 * See getopt(3)...
 */
int parseopt(struct options *opt, int argc, char **argv)
{
    char c;
    
    int  dport = 0,     // destination port
         xport = 1025,  // first source port
         yport = 65534, // last source port
         time  = 10000, // sleep time
         alarm = 40,    // alarm time
         window = 5840, // tcp window size
         frag = 0,      // fragment offset
         fragc = 1,     // fragment counter
         syn = 0,       // syn flooding only
         ack = 0,       // send acknowledgment packets
         fin = 0,       // send finalize packets
         rst = 0,       // send reset packets
         mlt = 0,       // multistage payload
         loop = 0,      // infinite loop
         children = 8,  // max processes to spawn
         ipver = 0,     // ip version
         verbose = 1;   // verbosity
    
    char dstaddr[MIN_ARG], // destination address
         *srcaddr  = NULL, // source address
         *device   = NULL, // network interface
         *firewall = NULL, // firewall rule
         *payload  = NULL, // payload file
         *mpayload = NULL; // multistage payload file
    
    struct addrinfo hints, *res=NULL;
    void *addr=NULL;
    
    int single = 1;
    
    // cleaning
    dstaddr[0] = '\0';
    memset(opt, 0, sizeof(struct options)); // clear opt
    memset(&hints, 0, sizeof(hints));
    
    char *shortopt = "d:p:x:y:ls:i:t:a:m:f:v:L:W:O:C:P:M:"; // short options
    
    while ((c = getopt (argc, argv, shortopt)) != -1)
        switch (c) {
        
            case 'd': // Destination
            
                // ip address
                if(regmatch(optarg, reg_ipv4_addr)) {
                    strncpy(dstaddr, optarg, MIN_ARG-1);
                    single = 1;
                }
                
                // dns name
                else if(getaddrinfo(optarg, NULL, &hints, &res) == 0) {
                
                    if (res->ai_family == AF_INET) { // IPv4
                        struct sockaddr_in *ipv4 = (struct sockaddr_in *) res->ai_addr;
                        addr = &(ipv4->sin_addr);
                        ipver=VER_IPV4;
                    }
                    
                    else if (res->ai_family == AF_INET6) { // IPv6
                        struct sockaddr_in6 *ipv6 = (struct sockaddr_in6 *)res->ai_addr;
                        addr = &(ipv6->sin6_addr);
                        ipver=VER_IPV6;
                    }

                    // convert the IP to a string and print it:
                    inet_ntop(res->ai_family, addr, dstaddr, MIN_ARG-1);
                    
                    single=1;
                    freeaddrinfo(res);
                }
                
                else {
                    usage(stderr, "\nArgument -d is invalid.\n");
                }
                
                break;
            
            case 'p': // Port
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -p is invalid.\n");
                dport = atoi(optarg);
                break;
            
            case 'x': // First source port
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -x is invalid.\n");
                xport = atoi(optarg);
                break;
            
            case 'y': // Last source port
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -y is invalid.\n");
                yport = atoi(optarg);
                break;
            
            case 'l': // Infinite loop
                loop = 1;
                break;
            
            case 's': // Source address
                if(!regmatch(optarg, reg_ipv4_addr))
                    usage(stderr, "\nArgument -s is invalid.\n");
                srcaddr = optarg;
                break;
            
            case 'i': // Interface
                device = optarg;
                break;
            
            case 't': // Sleep time
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -t is invalid.\n");
                time = atoi(optarg);
                break;
            
            case 'a': // Alarm time
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -a is invalid.\n");
                alarm = atoi(optarg);
                break;
            
            case 'm': // Max processes to spawn
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -m is invalid.\n");
                children = atoi(optarg);
                break;
            
            case 'f': // Firewall rule
                firewall = optarg;
                break;
            
            case 'v': // Verbosity level
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -v is invalid.\n");
                verbose = atoi(optarg);
                break;
            
            case 'L': // Level of interaction
                if(strstr(optarg, "r")) {
                    ack = 1; rst = 1;
                }
                else if(strstr(optarg, "f")) {
                    ack = 1; fin = 1;
                }
                else if(strstr(optarg, "a")) {
                    ack = 1;
                }
                else if(strstr(optarg, "s")) {
                    syn = 1;
                }
                else {
                    usage(stderr, "\nArgument -L is invalid.\n");
                }
                break;
            
            case 'W': // Window size
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -W is invalid.\n");
                window = atoi(optarg);
                break;
            
            case 'O': // Fragment offset
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -O is invalid.\n");
                frag = atoi(optarg);
                break;
            
            case 'C': // Fragment counter
                if(!regmatch(optarg, reg_number))
                    usage(stderr, "\nArgument -C is invalid.\n");
                fragc = atoi(optarg);
                break;
            
            case 'P': // Payload
                payload = optarg;
                break;
            
            case 'M': // Multistage payload
                mpayload = optarg;
                mlt = 1;
                break;
                            
            case '?':
            default:
                usage(stderr, NULL);
        }

    //check collected args and fill options struct
    if(dstaddr[0]=='\0' || dport<1)
        usage(stderr, NULL);
    else {
        strncpy(opt->dstaddr, dstaddr, MIN_ARG);
        opt->dport=dport;
    }
    
    if(srcaddr)
        strncpy(opt->srcaddr, srcaddr, MIN_ARG);
    
    if(time<0)
        time = 0;
    if(alarm<0)
        alarm = 0;
    if(xport>yport)
        usage(stderr, "\nAre you ok? Check -x and -y please ;)\n");
    if(children<0)
        children = 8;
    if(verbose<0)
        verbose = 0;
        
    if(firewall) {
        if(!strcmp("iptables", firewall)) {
            strncpy(opt->firewall[0], IPTABLES_ENABLE, MAX_ARG);
            strncpy(opt->firewall[1], IPTABLES_DISABLE, MAX_ARG);
        }
        else if(!strcmp("blackhole", firewall)) {
            strncpy(opt->firewall[0], BLACKHOLE_ENABLE, MAX_ARG);
            strncpy(opt->firewall[1], BLACKHOLE_DISABLE, MAX_ARG);
        }
        else
           usage(stderr, "\nI don't know this firewall\n");
    }

    // TODO Check ipversion

    opt->xport=xport;
    opt->yport=yport;
    opt->loop=loop;
    opt->time=time;
    opt->alarm=alarm;
    opt->window=window;
    opt->frag=frag;
    opt->fragc=fragc;
    opt->syn=syn;
    opt->ack=ack;
    opt->fin=fin;
    opt->rst=rst;
    opt->mlt=mlt;
    opt->children=children;
    opt->verbose=verbose;
    
    if(device)
        strncpy(opt->device, device, MIN_ARG);
    
    if(payload)
        getpayload(opt, payload);
    
    if(mpayload)
        getmultipayload(opt, mpayload);
    
    return 1;
}

#endif /* __ARGPARSER_H__ */

